/**
 * @file Ce fichier contient la classe de l'API du LDAP qui gère les opérations sur les groupes. Elle est destinée à être exportée pour être utilisée par les resolvers.
 * @author hawkspar
 * @memberof LDAP
 */

import { ldapConfig, categories } from '../internal/config';
import {Basics} from '../internal/basics';
import {Tools} from '../internal/tools';

//------------------------------------------------------------------------------------------------------------------------
// Types à exporter
//------------------------------------------------------------------------------------------------------------------------

/**
 * @memberof LDAP
 * @class partUserData
 * @desc Interface avec les données visibles au mpnde extérieur.
 * @var {string} uid - Identifiant utilisateur
 * @var {string} givenName - Prénom
 * @var {string} lastName - Nom
 * @var {string?} nickname - Surnom
 * @var {string?} gender - Sexe
 * @var {string?} photo - Bytestring de la photo de l'utilisateur
 * @var {string?} urlPhoto - Adresse de la photo de l'utilisateur
 * @var {string?} address - Adresse(s)
 * @var {string[]} admins - Liste des gid (group id, inclus section sportive, binet, PA...) dont l'utilisateur est admin ; pas forcément sous-liste de members
 * @var {string[]} speakers - Liste des gid dont l'utilisateur est porte-parole ; pas forcément sous-liste de members. Valeur par défaut est vide.
 * @var {string[]} members - Liste des gid dont l'utilisateur est membre. Valeur par défaut est vide.
 * @var {string[]} followers - Liste des gid dont l'utilisateur est sympathisant. Valeur par défaut est vide.
 * @var {string?} phone - Numéro(s) de téléphone
 * @var {string?} mail - Adresse(s) courriel
 * @var {string} birthdate - Date d'anniversaire
 * @var {string} nationality - Nationalité d'origine
 * @var {string} refNum - Matricule
 * @var {string?} password - Mot  de passe (jamais transmis en clair)
 * @var {string?} importDN - Domain Name d'importation pour les inscriptions en masse
 * @var {string?} login - Adresse de connexion à l'utilisateur directe 
 * @var {string?} directory - Zone du serveur où sont stockés les données utilisateur
 */
export class userData {
    uid: string;
    givenName: string;
    lastName: string;
    nickname?: string;
    gender?: 'M'|'F'|'U';
    photo?: string;
    urlPhoto?: string = '';
    address?: string;
    admins: string[] = [];
    speakers: string[] = [];
    members: string[] = [];
    followers: string[] = [];
    phone?: string;
    mail?: string;
    birthdate: string;
    nationality?: string;
    refNum: string;
    password?: string;
    importDN?: string;
    login?: string;
    directory?: string;
}

//------------------------------------------------------------------------------------------------------------------------
// Classe à exporter TBT
//------------------------------------------------------------------------------------------------------------------------

export class User {
    /**
     * @memberof LDAP
     * @class User
     * @classdesc Cette classe est une des deux classes exportables permettant de faire des opérations sur les utilisateurs.
     * @summary Constructeur vide.
     */
    constructor() {}
    
    /**
     * @memberof LDAP
     * @summary Fonction qui renvoit les infos du niveau voulues relatives à un utilisateur particulier.
     * @desc Cette fonction utilise {@link Tools.peek} avec l'interface {@link userData}.
     * @arg {string} uid - Identifiant de l'utilisateur, supposé valide.
     * @return {Promise(userData)} Informations recueillies au format {@link userData}.
     * @static
     * @async
     */
    static async peek(uid: string) : Promise<userData> {
        try         { return Tools.peek<userData>("user", uid, userData); }
        catch(err)  { throw "Error while peeking a user."; }
    }
    
    /**
     * @memberof LDAP
     * @summary Fonction qui retrouve les uid des paxs validant les critères de recherche. Utiliser {@link User.peek} au cas par cas après pour obtenir les vraies infos.
     * @desc Cette fonction utilise {@link Tools.search}.
     * @arg {userData} data - Dictionnaire contenant les données nécessaires à la recherche. Les valeurs sont celles entrées par l'utilisateur et sont par hypothèse
     * comme des sous-parties compactes des valeurs renvoyées. Tous les champs ci-dessous peuvent être indifféremment des listes (par exemple pour chercher un membre
     * de plusieurs groupes) ou des éléments isolés. Si un champ n'est pas pertinent, le mettre à '' ou undefined.
     * @return {Promise(string[])} gids des profils qui "match" les critères proposés.
     * @static
     * @async
     */
    static async search(data: userData) : Promise<string[]> {
        try         { return Tools.search("user", data); }
        catch(err)  { throw "Erreur lors de la recherche approximative d'un utilisateur."; }
    }

    /**
     * @memberof LDAP
     * @summary Fonction intermédiaire qui reformatte un fichier User en LDAP-friendly pour tous les champs éditables
     * @desc Cette fonction n'est pas asynchrone et modifie le dictionnaire passé par référence.
     * @arg {userData} data - Dictionnaire de données utilisateurs
     * @arg {dic} vals - Dictionnaire passé par référence pour être enrichi des données utilisateurs reformattées
     */
    static reformat(data: userData, vals : any) : void {
        // Majuscule sur les premières lettres
        vals[ldapConfig.user['givenName']] = data["givenName"][0].toUpperCase() + data["givenName"].toLowerCase().slice(1);
        vals[ldapConfig.user['lastName']] = data["lastName"][0].toUpperCase() + data["lastName"].toLowerCase().slice(1);

        // Génère une erreur si un champ n'est pas rempli
        for (let key_att of ["nickname","gender","photo","urlPhoto","phone","adress","mail","birthdate","nationality","refNum","importDN"]) {
            // Ecriture de toutes les valeurs uniques
            if (data[key_att] != undefined && data[key_att] != '')  vals[ldapConfig.user[key_att]]=data[key_att];
            else                                                    vals[ldapConfig.user[key_att]]='';
        }

        // Mot de passe crypté et salé
        vals[ldapConfig.user['password']] = "{CRYPT}"+data['password'];

        // Noms complets
        vals[ldapConfig.user['fullName']] = data["givenName"].toLowerCase() + " " + data["lastName"].toUpperCase();
        vals[ldapConfig.user['cleanFullName']] = vals[ldapConfig.user['fullName']];

        // Elimination des accents
        vals[ldapConfig.user['cleanFullNameEn']] = vals[ldapConfig.user['fullName']].normalize('NFD').replace(/[\u0300-\u036f]/g, "");
    }
    
    /**
     * @memberof LDAP
     * @summary Fonction qui créé un nouvel utilisateur dans le LDAP.
     * @desc Appelle {@link LDAP.add} bien sûr, mais aussi {@link Tools.add} pour gérer les groupes du nouvel utilisateur.
     * @arg {userData} data - Dictionnaire des informations utilisateurs. Des erreurs peuvent apparaître si tous les champs ne sont pas remplis.
     * Cette application permet de rejoindre des groupes en masse pour toute catégorie, à la fois façon Sigma et CAS.
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async create(data: userData): Promise<boolean> {
        // Calcul d'un dictionnaire d'ajout
        let vals = {};

        // uid de base généré à partir de nom et prénom, plus potentiellement promo et un offset
        // MEF mélange de Promise et de fonction standard
        try {
            Tools.generateUid(data['givenName'],data['lastName'],data['birthdate']).then(id => { vals[ldapConfig.user.uid]=id; } );
        }
        catch(err) {
            throw "Erreur lors de la génération d'un hruid pour un nouvel utilisateur.";
        }

        let uid : string = vals[ldapConfig.user.uid];

        // Réécriture données utilisateurs en mode LDAP-friendly
        User.reformat(data, vals);

        // Appel à la fonction de base
        if (!await Basics.add("user", vals)) throw "Erreur de l'ajout de la feuille à l'arbre utilisateur.";
        
        // Certains champs nécessitent de petits calculs
        let vals2 = {}

        // Adressage root
        if (data['members'].includes("on_platal")) { vals2[ldapConfig.user['login']] = "/bin/bash"; }
        else  { vals2[ldapConfig.user['login']] = "/sbin/nologin"; }
        
        // Stockage machine ; dépend du prénom
        vals2[ldapConfig.user['directory']] = '/hosting/users/' + vals['cleanFullNameEn'][0];

        // Valeur nécessaire ASKIP mais inutile
        vals2[ldapConfig.user['const']] ='5000';
        // Création identifiant unique numérique aléatoire        
        vals2[ldapConfig.user['idNumber']] = await Tools.generateId(ldapConfig.user['idNumber'], "user");

        // Ajout LDAP
        try {
            if (!await Basics.change("user", uid, "add", vals2)) throw "Erreur lors de l'ajout des valeurs calculées à la feuille du nouvel utilisateur.";
        }
        catch(err) { throw "Erreur lors de la génération d'un id numérique pour un nouvel utilisateur."; }

        ["posixAccount", "shadowAccount", "brUser"].forEach(cst => {
            let tmp = { [ldapConfig.user['classes']] : cst };
            Basics.change("user", uid, "add", tmp).then(res => {
                if (!res) throw "Erreur lors de l'ajout d'une valeur constante à la feuille du nouvel utilisateur.";
            });
        });

        // Ajout dans les groupes à la catégorie voulue
        for (let cat of categories) {
            for (let gid of data[cat]) Tools.add(uid, gid, cat).then(res => {
                if (!res) throw "Erreur de l'ajout d'un membre au nouveau groupe.";
            });
        }
        return true;
    }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de suppression TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @memberof LDAP
     * @summary Fonction qui supprime un utilisateur du LDAP.
     * @desc Cette fonction commence par gérer les groupes du membre puis le supprime entièrement.
     * Appelle {@link LDAP.clear} bien sûr, mais aussi {@link Tools.remove} pour gérer les groupes de l'utilisateur sortant.
     * @arg {string} uid - uid de la victime
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async delete(uid: string): Promise<boolean> {
        try {
            // Gestion des groupes de l'utilisateur d'abord
            let profil = await User.peek(uid);
            for (let cat of categories) profil[ldapConfig.user[cat]].forEach(gid => Tools.remove(uid, gid, cat));
        }
        catch(err) {
            throw "Erreur lors de l'obtention des informations de l'utilisateur à supprimer.";
        }
        // Elimination pure et simple
        if (!Basics.clear("user", uid)) throw "Erreur lors de la suppression de l'utilisateur.";
        return true;
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui édite un utilisateur existant dans le LDAP.
     * @desc Appelle simplement {@link User.delete} suivi de {@link User.create}. Sans effet sur les groupes de l'utilisateur concerné.
     * Cette méthode ne permet pas de rejoindre aucun groupe, mais elle permet de changer son mot de passe, image, ou même son nom.
     * @arg {userData} data - Dictionnaire des informations utilisateurs
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async edit(data : userData) : Promise<boolean> {
        try {
            let vals = {};
            User.reformat(data, vals);
            // Modification par effet de bord
            return Basics.change("user", data["uid"], "replace", vals);
        }
        catch(err) { throw "Erreur lors de la modification d'un utilisateur."; }
    }
}